#include <cmath>
#include <cstdlib>
#include <algorithm>
#include <numeric>
#include <iostream>

#ifdef _DUMP_DATA
#include <fstream>
#endif

#include "homography2d.h"

// Two ways are equivalent, just for legacy purpose.
// If there is no <rknave/clapack_wrapper.h>,
// Download clapack.h and f2c.h from web, uncomment the next line.
#define _RANSAC_USE_LAPACK_
#ifdef _RANSAC_USE_LAPACK_
#ifdef __cplusplus 
extern "C" {
#endif
#include "F2CLIBS/f2c.h" 
#include "clapack.h"
#ifdef __cplusplus
}
#endif
#else
#include <rknave/clapack_wrapper.h>
#endif

int homography2d(ProjectiveTransformation::WarpModel w, double *pts, const int rows, const int npts, double H[9])
{
	if(w == ProjectiveTransformation::Homography)
		return homography2d_homography(pts, rows, npts, H);
	else if(w == ProjectiveTransformation::Affine)
		return homography2d_affine(pts, rows, npts, H);
	else if(w == ProjectiveTransformation::Similarity)
		return homography2d_similarity(pts, rows, npts, H);
	return 0;
}

/*
% HOMOGRAPHY2D - computes 2D homography
%
% Usage:   int homography2d(x, rows, npts, H)
%
% Inputs:
%          x  - x = [x1; x2], column major
% Outputs:
%          H - the 3x3 homography such that x2 = H*x1
%
% This code follows the normalized Direct Linear Transformation 
% algorithm given by Hartley and Zisserman "Multiple View Geometry in
% Computer Vision (2nd edition)" p89.
*/
/*
template <> int homography2d<ProjectiveTransformation::Homography>
(double *pts, const int rows, const int npts, double H[9])
*/
int homography2d_homography(double *pts, const int rows, const int npts, double H[9])
{
	double *x1 = pts;
	double *x2 = pts+rows/2;
    double T1[9];
	double T2[9];
	int n1 = normalize2dpts(x1, rows, npts, x1, T1);
	int n2 = normalize2dpts(x2, rows, npts, x2, T2);
	
	double *A = new double[npts*18];
    // Memory layout: column-major (LAPACK likes it)
	for (int i = 0; i < npts; i++)
	{
		double x = x2[i*rows  ];
		double y = x2[i*rows+1];
		double w = x2[i*rows+2];
		double X = x1[i*rows  ];
		double Y = x1[i*rows+1];
		double W = x1[i*rows+2];

		int ibase = i*18;
		A[ibase   ] = 0.0;
		A[ibase+1 ] = 0.0;
		A[ibase+2 ] = 0.0;
		
		A[ibase+3 ] = -w*X;
		A[ibase+4 ] = -w*Y;
		A[ibase+5 ] = -w*W;
        
		A[ibase+6 ] = y*X;
		A[ibase+7 ] = y*Y;
		A[ibase+8 ] = y*W;

		A[ibase+9 ] = w*X;
		A[ibase+10] = w*Y;
		A[ibase+11] = w*W;
		
		A[ibase+12] = 0.0;
		A[ibase+13] = 0.0;
		A[ibase+14] = 0.0;
        
		A[ibase+15] = -x*X;
		A[ibase+16] = -x*Y;
		A[ibase+17] = -x*W;
	}

	char JOBU	= 'A';
    char JOBVT	= 'N';
	integer M = 9;
    integer N = npts*2;
    integer LDA = M;
    integer LDU = M;
    integer LDVT = N;
    integer LWORK =100+2*npts;
    integer INFO;
	double *wk = new double[LWORK];
	double s[9];
	double uu[81];
	//double *vt = new double[9*npts*npts];
    double *vt = 0;

	dgesvd_( &JOBU, &JOBVT, &M, &N, A, &LDA, s, uu, 
          &LDU, vt, &LDVT, wk, &LWORK, &INFO);

	// Denormalise
    //H = T2\H*T1;
	double *h = &(uu[72]);

    denormalize_transposed(h, T1, T2, H);

	//delete []vt;
	delete []wk;
	delete []A;

	return INFO;
}

/*
template <> int homography2d<ProjectiveTransformation::Affine>
(double *pts, const int rows, const int npts, double H[9])
*/
int homography2d_affine(double *pts, const int rows, const int npts, double H[9])
{
    double *x1 = pts;
	double *x2 = pts+rows/2;
    double T1[9];
	double T2[9];
	int n1 = normalize2dpts(x1, rows, npts, x1, T1);
	int n2 = normalize2dpts(x2, rows, npts, x2, T2);
    
    double *A = new double[npts*4];
    // Memory layout: column-major (LAPACK likes it)
	for (int i = 0; i < npts; i++)
	{
        int irows = i*rows;
        A[i*4  ] = x1[irows  ];
        A[i*4+1] = x1[irows+1];
        A[i*4+2] = x2[irows  ];
        A[i*4+3] = x2[irows+1];        
    }
    char JOBU	= 'A';
    char JOBVT	= 'N';
	integer M = 4;
    integer N = npts;
    integer LDA = M;
    integer LDU = M;
    integer LDVT = N;
    integer LWORK = 50+npts;
    integer INFO;
	double *wk = new double[LWORK];
	double s[4];
	double uu[16];
	//double *vt = new double[9*npts*npts];
    double *vt = 0;

	dgesvd_( &JOBU, &JOBVT, &M, &N, A, &LDA, s, uu, 
          &LDU, vt, &LDVT, wk, &LWORK, &INFO);

	// Denormalise
    //H = T2\H*T1;

	double h[9];
    //[ (-h2*h5+h6*h1)/(-h0*h5+h4*h1),  (h2*h4-h6*h0)/(-h0*h5+h4*h1) 0]
    //[ (-h3*h5+h7*h1)/(-h0*h5+h4*h1),  (h3*h4-h7*h0)/(-h0*h5+h4*h1) 0]
    //[ 0                               0                            1]
    h[0] = uu[6]*uu[1]-uu[2]*uu[5];
    h[3] = uu[7]*uu[1]-uu[3]*uu[5];
    h[6] = 0.0;

    h[1] = uu[2]*uu[4]-uu[6]*uu[0];
    h[4] = uu[3]*uu[4]-uu[7]*uu[0];
    h[7] = 0.0;

    h[2] = 0.0;
    h[5] = 0.0;
    h[8] = uu[4]*uu[1]-uu[0]*uu[5];

	denormalize_transposed(h, T1, T2, H);

    delete []wk;
    delete []A;
    return INFO;
}

// for Similarity 
static inline double ddist2(const double *p1, const double *p2, const int dim)
{
    double s = 0.0;
    for (int i = 0; i < dim; i++)
        s += (p1[i] - p2[i]) * (p1[i] - p2[i]);
    return s;
}
static inline void addvec(const double *pt, const int dim, double *acc)
{
    for (int i = 0; i < dim; i++)
        acc[i] += pt[i];
}
static inline double sigma2(const double *pt, const int dim, const int step, const int n, const double *mv)
{
    double s = 0.0;
    if ( n <= 0 )
        return s;
    for (int i = 0; i < n; i++){
        s += ddist2(pt+i*step, mv, dim);
    }
    return s / double(n);
}
static inline void meanvec(const double *pt, const int dim, const int step, const int n, double *mv)
{
	int i;
    std::fill(mv, mv+dim, 0.0);
    if (n <= 0)
        return;
    for (i = 0; i < n; i++){
        addvec(pt+i*step, dim, mv);
    }
    for (i = 0; i < dim; i++)
        mv[i] /= double(n);
}

static inline void addcov(const double *x, const double *y, const int dim,
                          const double *mx, const double *my, double *cov, const double c = 1.0)
{
    for (int i = 0; i < dim; i++)
        for (int j = 0; j < dim; j++)
            cov[i*dim+j] += c * (y[i] - my[i]) * (x[j] - mx[j]);
}


static inline void compute_cov(const double *x, const double *y, const int dim, const int step, const int n,
                        const double *mx, const double *my, double *cov)
{
	int i;
    std::fill(cov, cov+dim*dim, 0.0);
    if (n <= 0)
        return;
    for (i = 0; i < n; i++)
        addcov(x+i*step, y+i*step, dim, mx, my, cov);
    
    for (i = 0; i < dim*dim; i++)
        cov[i] /= double(n);
}

static inline double determinant2(const double *mat)
{
    return (mat[0]*mat[3] - mat[1]*mat[2]);
}

static inline void matvec(const double *mat, const double *x, double *y, const int m, const int n)
{
    std::fill(y, y+m, 0.0);
    for (int i = 0; i < n; i++){
        for (int j = 0; j < m; j++)
            y[j] += mat[i*m + j] * x[i];
    }
}

// Similarity transformation:
// Reference: Shinji Umeyama, Least-Sequares Estimation of Transformation
// Parameters Between Two Point Patterns, pp376-380, PAMI vol.13, no.4 apr 1991.
/*
template <> int homography2d<ProjectiveTransformation::Similarity>
(double *pts, const int rows, const int npts, double H[9])
*/
int homography2d_similarity(double *pts, const int rows, const int npts, double H[9])
{
    double *x = pts;
	double *y = pts+rows/2;
    const int dim = 2;
    double mx[dim];
    double my[dim];
    double covxy[dim*dim];
    
    meanvec(x, dim, rows, npts, mx);
    meanvec(y, dim, rows, npts, my);
    
    double sigx2 = sigma2(x, dim, rows, npts, mx);
    double sigy2 = sigma2(y, dim, rows, npts, my);
    
    compute_cov(x, y, dim, rows, npts, mx, my, covxy);

	int i;
    const int LWK = 32;
    char JOBU	= 'A';
    char JOBVT  = 'A';
	integer M = dim;
    integer N = dim;
    integer LDA = M;
    integer LDU = M;
    integer LDVT = N;
    integer LWORK = LWK;
    integer INFO;
	double wk[LWK];
	double s[dim];
	double uu[dim*dim];
	double vt[dim*dim];
    
    dgesvd_( &JOBU, &JOBVT, &M, &N, covxy, &LDA, s, uu, 
          &LDU, vt, &LDVT, wk, &LWORK, &INFO);

    double chirality = 1.0;
    if ( determinant2(uu) * determinant2(vt) < 0.0 )
        chirality = -1.0;

    double c = 0.0;
    double R[dim*dim];
    std::fill(R, R+dim*dim, 0.0);
    double zerovec[dim];
    std::fill(zerovec, zerovec+dim, 0.0);
    for (i = 0; i < dim; i++){
        double chi = (i==dim-1)? chirality:1.0;
        c += s[i]*chi;
        addcov(uu+dim*i, vt+dim*i, dim, zerovec, zerovec, R, chi);
    }
    c /= sigx2;

    double t[dim];
    matvec(R, mx, t, dim, dim);
    for (i = 0; i < dim; i++)
        t[i] *= -c;
    addvec(my, dim, t);
    
    H[0] = c*R[0];
    H[1] = c*R[1];
    H[2] = 0.0;
    
    H[3] = c*R[2];
    H[4] = c*R[3];
    H[5] = 0.0;

    H[6] = t[0];
    H[7] = t[1];
    H[8] = 1.0;

    return 0;
}


/*
% NORMALISE2DPTS - normalises 2D homogeneous points
%
% Function translates and normalises a set of 2D homogeneous points 
% so that their centroid is at the origin and their mean distance from 
% the origin is sqrt(2).
%
% Usage:   int normalise2dpts(pts, rows, npts, newpts, T)
%
% Argument:
%   pts -  3xN array of 2D homogeneous coordinates
%
% Returns:
%   newpts -  3xN array of transformed 2D homogeneous coordinates.  The
%             scaling parameter is normalized to 1 unless the point is at
%             infinity. 
%   T      -  The 3x3 transformation matrix, newpts = T*pts
%           
% If there are some points at infinity the normalization transform
% is calculated using just the finite points.  Being a scaling and
% translating transform this will not affect the points at infinity.
*/

// pts & newpts: In-place is allowed.
int normalize2dpts(double *pts, const int rows, const int npts, double *newpts, double T[9])
{
	// Find the indices of the points that are not at infinity
	int i;
	int n = 0;
	double c0 = 0.0;
	double c1 = 0.0;
	for (i = 0; i < npts; i++)
	{
		if(pts[i*rows+2] > RansacSetting::dbl_eps_2 || pts[i*rows+2] < -RansacSetting::dbl_eps_2)
		{
			c0 += newpts[i*rows  ] = pts[i*rows  ] / pts[i*rows+2];
            c1 += newpts[i*rows+1] = pts[i*rows+1] / pts[i*rows+2];
			newpts[i*rows+2] = 1.0;
			++n;
		}
		else
		{
			newpts[i*rows  ] = pts[i*rows  ];
			newpts[i*rows+1] = pts[i*rows+1];
			newpts[i*rows+2] = pts[i*rows+2];
		}
	}

	if (n == 0)
	{
		T[0] = 1.0;
		T[4] = 1.0;
		T[6] = 0.0;
		T[7] = 0.0;
        return n;
	}

	c0 /= n;
	c1 /= n;
	double meandist = 0.0;
    for (i = 0; i < npts; i++)
	{
		if (newpts[i*rows+2] > 0.99)
		{
			double d0 = newpts[i*rows  ] - c0;
			double d1 = newpts[i*rows+1] - c1;
			meandist += sqrt(d0*d0 + d1*d1);
		}
	}

	meandist /= n;

	double scale = RansacSetting::dbl_sqrt_2 / (meandist + RansacSetting::dbl_eps_2);

    if (scale <= RansacSetting::dbl_eps_2)
		scale = RansacSetting::dbl_eps_2;

	for (i = 0; i < npts; i++)
	{
        newpts[i*rows  ] = (newpts[i*rows  ] - c0)*scale;
		newpts[i*rows+1] = (newpts[i*rows+1] - c1)*scale;
	}
	
	/*
	T = [scale   0   -scale*c(1)
         0     scale -scale*c(2)
         0       0      1      ];
		 */
	T[0] = scale;
//	T[1] = 0.0;
//	T[2] = 0.0;
//	T[3] = 0.0;
	T[4] = scale;
//	T[5] = 0.0;
	T[6] = -scale*c0;
	T[7] = -scale*c1;
//	T[8] = 1.0;
	
	return n;
}

// H = T2\H^T*T1;
void denormalize_transposed(const double h[9], const double T1[9], const double T2[9], double H[9])
{
    double _T2 = 1.0 / T2[0];
    double T1_T2 = T1[0] * _T2;
    double h0_T26xh6 = h[0]-T2[6]*h[6];
    double h1_T26xh7 = h[1]-T2[6]*h[7];
    double h3_T27xh6 = h[3]-T2[7]*h[6];
    double h4_T27xh7 = h[4]-T2[7]*h[7];
	H[0] = h0_T26xh6 * T1_T2;//T2[0]*T1[0];
	H[1] = h3_T27xh6 * T1_T2;//T2[0]*T1[0];
	H[2] = h[6]*T1[0];
	H[3] = h1_T26xh7 * T1_T2;//T2[0]*T1[0];
	H[4] = h4_T27xh7 * T1_T2;//T2[0]*T1[0];
	H[5] = h[7]*T1[0];
	H[6] = ( h0_T26xh6*T1[6]+h1_T26xh7*T1[7]+(h[2]-T2[6]*h[8]) )*_T2;
	H[7] = ( h3_T27xh6*T1[6]+h4_T27xh7*T1[7]+(h[5]-T2[7]*h[8]) )*_T2;
	H[8] = h[6]*T1[6]+h[7]*T1[7]+h[8];
}

// H = T2\H*T1; In-place allowed.
void denormalize(double H[9], const double T1[9], const double T2[9])
{
	double h0 = H[0];
	double h1 = H[3];
	double h2 = H[6];
	double h3 = H[1];
	double h4 = H[4];
	double h5 = H[7];
	double h6 = H[2];
	double h7 = H[5];
	double h8 = H[8];
	
    double _T2 = 1.0 / T2[0];
    double T1_T2 = T1[0] * _T2;
    H[0] = (h0-T2[6]*h6)*T1_T2;//T2[0]*T1[0];
    H[1] = (h3-T2[7]*h6)*T1_T2;//T2[0]*T1[0];
    H[2] = h6*T1[0];
    H[3] = (h1-T2[6]*h7)*T1_T2;//T2[0]*T1[0];
    H[4] = (h4-T2[7]*h7)*T1_T2;//T2[0]*T1[0];
    H[5] = h7*T1[0];
    H[6] = ((h0-T2[6]*h6)*T1[6]+(h1-T2[6]*h7)*T1[7]+(h2-T2[6]*h8))*_T2;
    H[7] = ((h3-T2[7]*h6)*T1[6]+(h4-T2[7]*h7)*T1[7]+(h5-T2[7]*h8))*_T2;
    H[8] = h6*T1[6]+h7*T1[7]+h8;
    double _H8 = 1.0 / H[8];
    for (int i = 0; i < 9; ++i)
        H[i] *= _H8;
}

static inline double cosvec2(const double* v0, const double *v1, double &denorm2)
{
    denorm2 = sqrt((v0[0]*v0[0]+v0[1]*v0[1])*(v1[0]*v1[0]+v1[1]*v1[1]));
    return (v0[0]*v1[0]+v0[1]*v1[1]);
}

// Check if matching is parallel?
template<> inline bool isnonparallel<ParallelCheck::Level0>
(const double *nx, const int rows, const int *inliers, const int ninliers,
 const double costhresh, const double voffset, const int ntrials)
{
    return false;
}

template<> bool isnonparallel<ParallelCheck::Level1>
(const double *nx, const int rows, const int *inliers, const int ninliers,
 const double costhresh, const double voffset, const int ntrials)
{
    if (ninliers < 4)
        return true;

    int halfrows = rows/2;
    int nonpara = 0;
    for (int i = 0; i < ntrials; i++){
        int rnd0 = 0;
        int rnd1 = 0;
        while (rnd0 == rnd1 || rnd0 >= ntrials || rnd1 >= ntrials) {
            rnd0 = static_cast<int>((static_cast<double>(rand())/RAND_MAX) * ninliers);
            rnd1 = static_cast<int>((static_cast<double>(rand())/RAND_MAX) * ninliers);
        }
        rnd0 = inliers[rnd0];
        rnd1 = inliers[rnd1];
        double v0[2];
        double v1[2];
        v0[0] = nx[rnd0*rows] - nx[rnd0*rows+halfrows];
        v0[1] = nx[rnd0*rows+1] - nx[rnd0*rows+halfrows+1] - voffset;
        v1[0] = nx[rnd1*rows] - nx[rnd1*rows+halfrows];
        v1[1] = nx[rnd1*rows+1] - nx[rnd1*rows+halfrows+1] - voffset;
        double denorm2;
        double cosv2 = cosvec2(v0, v1, denorm2);
        if (fabs(cosv2) < costhresh*denorm2)
            ++nonpara;
    }
    if (double(nonpara) > double(ntrials) * 0.3)
        return true;
    else
        return false;
}

/*
% Function to evaluate the symmetric transfer error of a homography with
% respect to a set of matched points as needed by RANSAC.
% function int homogdist2d(H, x, rows, npts, t, inliers);
% Side affect: x is changed by function hnormalize();
*/
int homogdist2d(const double H[9], const double *x, const int rows, const int npts, const double t, int *inliers)
{
    double A[9];
	int i;
#if 0
	double detH = mat_inverse3x3(H, invH);
    
    if (detH == dbl_eps_2)
        return 0;
#endif
    std::copy(H, H+9, A);

	double *nx = new double[rows*npts];
	double *Hx = new double[rows*npts];

    std::copy(x, x+rows*npts, Hx);

	// Matrix multiplication
	for (i = 0; i < npts; i++)
	{
		int irows = i*rows;
	//	int irows_2 = irows + rows/2;
		Hx[i*rows  ] = H[0]*x[irows]+H[3]*x[irows+1]+H[6]*x[irows+2];
		Hx[i*rows+1] = H[1]*x[irows]+H[4]*x[irows+1]+H[7]*x[irows+2];
		Hx[i*rows+2] = H[2]*x[irows]+H[5]*x[irows+1]+H[8]*x[irows+2];
#if 0
        Hx[i*rows+3] = invH[0]*x[irows_2]+invH[3]*x[irows_2+1]+invH[6]*x[irows_2+2];
		Hx[i*rows+4] = invH[1]*x[irows_2]+invH[4]*x[irows_2+1]+invH[7]*x[irows_2+2];
		Hx[i*rows+5] = invH[2]*x[irows_2]+invH[5]*x[irows_2+1]+invH[8]*x[irows_2+2];
#endif
	}

    integer N = 3;
    integer NRHS = npts;
    integer LDA = 3;
    integer LDB = rows;
    integer IPIV[3];
    integer INFO;

    double *B = Hx+rows/2;
	
    dgesv_( &N, &NRHS, A, &LDA, IPIV, B, &LDB, &INFO );

	hnormalize(x, rows, npts, nx);
	hnormalize(x+rows/2, rows, npts, nx+rows/2);
	hnormalize(Hx, rows, npts, Hx);
    hnormalize(Hx+rows/2, rows, npts, Hx+rows/2);
    
	int n = 0;
	for (i = 0; i < npts; i++)
	{
		int irows = i*rows;
		int irows_2 = irows + rows/2;
		double ste = ((nx[irows] - Hx[irows_2])*(nx[irows] - Hx[irows_2]) +
                      (nx[irows+1] - Hx[irows_2+1])*(nx[irows+1] - Hx[irows_2+1]) +
			          (nx[irows_2] - Hx[irows])*(nx[irows_2] - Hx[irows]) +
                      (nx[irows_2+1] - Hx[irows+1])*(nx[irows_2+1] - Hx[irows+1]));

        if (  ste < t )
		{
            inliers[n++] = i;
		}
	}
    
	// careful::modified
    if (isnonparallel<ParallelCheck::level>(nx, rows, inliers, n, 0.85, 2.0, 100))
        n = 0;

	delete []nx;
	delete []Hx;

	return n;
}



static double mat_inverse3x3(const double h[9], double invh[9])
{
	double det = -h[0]*h[4]*h[8]+h[0]*h[7]*h[5]+h[1]*h[3]*h[8]-h[1]*h[6]*h[5]-h[2]*h[3]*h[7]+h[2]*h[6]*h[4];

	if ( det >= -RansacSetting::dbl_eps_2 && det <= RansacSetting::dbl_eps_2)
		det = RansacSetting::dbl_eps_2;

	invh[0] = -(h[4]*h[8]-h[7]*h[5])/det;
	invh[1] = (h[1]*h[8]-h[7]*h[2])/det;
	invh[2] = -(h[1]*h[5]-h[4]*h[2])/det;

    invh[3] = (h[3]*h[8]-h[6]*h[5])/det;
	invh[4] = (-h[0]*h[8]+h[6]*h[2])/det;
	invh[5] = -(-h[0]*h[5]+h[3]*h[2])/det;
	
	invh[6] =  -(h[3]*h[7]-h[6]*h[4])/det;
	invh[7] = -(-h[0]*h[7]+h[6]*h[1])/det;
	invh[8] = (-h[0]*h[4]+h[3]*h[1])/det;

	return det;
}


/*
% HNORMALISE - Normalises array of homogeneous coordinates to a scale of 1
%
% Usage:  hnormalise(x, rows, npts, nx)
%
% Input:
%         x  - an Nxnpts array of homogeneous coordinates.
%
% Outputs:
%         nx - an Nxnpts array of homogeneous coordinates rescaled so
%              that the scale values nx(3,:) are all 1.
%
% Note that any homogeneous coordinates at infinity (having a scale value of
% 0) are left unchanged.
*/
// In-place is allowed: Side affect: x is changed.
void hnormalize(const double *x, const int rows, const int npts, double *nx)
{
	for (int i = 0; i < npts; i++)
	{
        if (x[i*rows+2] > RansacSetting::dbl_eps_2 || x[i*rows+2] < -RansacSetting::dbl_eps_2)
		{
			nx[i*rows  ] = x[i*rows  ] / x[i*rows+2];
            nx[i*rows+1] = x[i*rows+1] / x[i*rows+2];
			nx[i*rows+2] = 1.0;
		}
		else
		{
			nx[i*rows  ] = x[i*rows  ];
            nx[i*rows+1] = x[i*rows+1];
			nx[i*rows+2] = x[i*rows+2];
		}
	}
}

// for Similarity warp model
/*
template < > bool isdegenerate <ProjectiveTransformation::Similarity, DegeneracyCheck::Level0> 
(const double *x, const int m)
*/
bool isdegenerate_similarity_0(const double *x, const int m)
{
    const double *y = x + m/2;
    if ( (fabs(x[0] - x[m]) < RansacSetting::det_eps_2 && fabs(x[1] - x[m+1]) < RansacSetting::det_eps_2) ||
         (fabs(y[0] - y[m]) < RansacSetting::det_eps_2 && fabs(y[1] - y[m+1]) < RansacSetting::det_eps_2)  )
        return true;
    else
        return false;
}

// initial version: check angle == 0 ?
// for Affine warp model
/*
template < > bool isdegenerate <ProjectiveTransformation::Affine, DegeneracyCheck::Level0> 
(const double *x, const int m)
*/
bool isdegenerate_affine_0(const double *x, const int m)
{
    int m2 = m+m;
	const double *y = x + m/2;
	return (iscolinear(x  , x+m , x+m2) || iscolinear(y  , y+m , y+m2));
}

// for Homography
/*
template < > bool isdegenerate <ProjectiveTransformation::Homography, DegeneracyCheck::Level0> 
(const double *x, const int m)
*/bool isdegenerate_homography_0(const double *x, const int m)
{
	int m2 = m+m;
	int m3 = m2+m;
	const double *y = x + m/2;
	return (iscolinear(x  , x+m , x+m2) ||
			iscolinear(x  , x+m , x+m3) ||
			iscolinear(x  , x+m2, x+m3) ||
			iscolinear(x+m, x+m2, x+m3) ||
			iscolinear(y  , y+m , y+m2) ||
			iscolinear(y  , y+m , y+m3) ||
			iscolinear(y  , y+m2, y+m3) ||
			iscolinear(y+m, y+m2, y+m3) );
}

static inline bool iscolinear(const double *p0, const double *p1, const double *p2)
{
	double a0 = p1[0] - p0[0];
	double a1 = p1[1] - p0[1];
	double a2 = p1[2] - p0[2];

	double b0 = p2[0] - p0[0];
	double b1 = p2[1] - p0[1];
	double b2 = p2[2] - p0[2]; 

	double c0 = a1*b2 - a2*b1;
	double c1 = a2*b0 - a0*b2;
	double c2 = a0*b1 - a1*b0;

	return (c0*c0 + c1*c1 + c2*c2 < RansacSetting::dbl_eps_2);
}


// check det == 0 and sign is different (same triangle order)
/*
template < > bool isdegenerate <ProjectiveTransformation::Homography, DegeneracyCheck::Level1> 
(const double *x, const int m)
*/
bool isdegenerate_homography_1(const double *x, const int m)
{
	int m2 = m+m;
	int m3 = m2+m;
	const double *y = x + m/2;
	return (signofdet3(x  , x+m , x+m2) * signofdet3(y  , y+m , y+m2) <= 0 ||
			signofdet3(x  , x+m , x+m3) * signofdet3(y  , y+m , y+m3) <= 0 ||
			signofdet3(x  , x+m2, x+m3) * signofdet3(y  , y+m2, y+m3) <= 0 ||
			signofdet3(x+m, x+m2, x+m3) * signofdet3(y+m, y+m2, y+m3) <= 0 );
}

// for Affine warp model
/*
template < > bool isdegenerate <ProjectiveTransformation::Affine, DegeneracyCheck::Level1>
(const double *x, const int m)
*/
bool isdegenerate_affine_1(const double *x, const int m)
{
    int m2 = m+m;
	const double *y = x + m/2;
	return (signofdet3(x  , x+m , x+m2) * signofdet3(y  , y+m , y+m2) <= 0);
}

static inline char signofdet3(const double *p0, const double *p1, const double *p2)
{
	
	double det = determinant3(p0, p1, p2);
    if ( det > RansacSetting::det_eps_2 )
        return 1;
    else if( det < -RansacSetting::det_eps_2 )
        return -1;
    else
        return 0;
}


// area ratio cannot be too large or too small.
/*
template < > bool isdegenerate <ProjectiveTransformation::Homography, DegeneracyCheck::Level2 >
(const double *x, const int m)
*/
bool isdegenerate_homography_2(const double *x, const int m)
{
	int m2 = m+m;
	int m3 = m2+m;
	const double *y = x + m/2;

    double detx, dety;
    detx = determinant3(x  , x+m , x+m2);
    dety = determinant3(y  , y+m , y+m2);
    if (detx * dety <= 0 || 
        fabs(detx) < fabs(dety) * RansacSetting::det_small_ratio || 
        fabs(detx) > fabs(dety) * RansacSetting::det_large_ratio )
        return true;

    detx = determinant3(x  , x+m , x+m3);
    dety = determinant3(y  , y+m , y+m3);
    if (detx * dety <= 0 || 
        fabs(detx) < fabs(dety) * RansacSetting::det_small_ratio || 
        fabs(detx) > fabs(dety) * RansacSetting::det_large_ratio )
        return true;

    detx = determinant3(x  , x+m2, x+m3);
    dety = determinant3(y  , y+m2, y+m3);
    if (detx * dety <= 0 || 
        fabs(detx) < fabs(dety) * RansacSetting::det_small_ratio || 
        fabs(detx) > fabs(dety) * RansacSetting::det_large_ratio )
        return true;

    detx = determinant3(x+m, x+m2, x+m3);
    dety = determinant3(y+m, y+m2, y+m3);
    if (detx * dety <= 0 || 
        fabs(detx) < fabs(dety) * RansacSetting::det_small_ratio || 
        fabs(detx) > fabs(dety) * RansacSetting::det_large_ratio )
        return true;

    return false;
}

// for Affine warp model
/*
template < > bool isdegenerate <ProjectiveTransformation::Affine, DegeneracyCheck::Level2>
(const double *x, const int m)
*/
bool isdegenerate_affine_2(const double *x, const int m)
{
    int m2 = m+m;
	const double *y = x + m/2;
    double detx = determinant3(x  , x+m , x+m2);
    double dety = determinant3(y  , y+m , y+m2);
    if (detx * dety <= 0 || 
        fabs(detx) < fabs(dety) * RansacSetting::det_small_ratio || 
        fabs(detx) > fabs(dety) * RansacSetting::det_large_ratio )
        return true;
    else
        return false;
}

static inline double determinant3(const double *p0, const double *p1, const double *p2)
{
	
	double det = p0[0]*p1[1]*p2[2] + p0[1]*p1[2]*p2[0] + p0[2]*p1[0]*p2[1]
			   - p0[0]*p1[2]*p2[1] - p0[1]*p1[0]*p2[2] - p0[2]*p1[1]*p2[0];
    return det;
}

// Check for Homography model using parallel constraints
/*
template < > bool isdegenerate <ProjectiveTransformation::Homography, DegeneracyCheck::Level3 >
(const double *x, const int m)
*/
bool isdegenerate_homography_3(const double *x, const int m)
{
    // TODO
    return false;
}

// Check for Affine model using parallel constraints
/*
template < > bool isdegenerate <ProjectiveTransformation::Affine, DegeneracyCheck::Level3>
(const double *x, const int m)
*/
bool isdegenerate_affine_3(const double *x, const int m)
{
    // TODO
    return false;
}

// Check if the homograpgy is valid using 4 canonic points. If it
// is too skewed, then discard it.
bool invalidhomography(const double H[9])
{
    // homography checking
    const double canonic4points[12] = {0.0, 0.0, 1.0, 1.0, 0.0, 1.0, 1.0, 1.0, 1.0, 0.0, 1.0, 1.0};
    double project4points[12];
    double ptpairs[24];

    // Matrix multiplication
    for (int i = 0; i < 4; i++)
    {
        project4points[i*3  ] = H[0]*canonic4points[i*3]+H[3]*canonic4points[i*3+1]+H[6]*canonic4points[i*3+2];
        project4points[i*3+1] = H[1]*canonic4points[i*3]+H[4]*canonic4points[i*3+1]+H[7]*canonic4points[i*3+2];
        project4points[i*3+2] = H[2]*canonic4points[i*3]+H[5]*canonic4points[i*3+1]+H[8]*canonic4points[i*3+2];
        // normalize it, may not need it, but in case....
        if(project4points[i*3+2] > RansacSetting::dbl_eps_2 || project4points[i*3+2] < -RansacSetting::dbl_eps_2){
            project4points[i*3  ] /= project4points[i*3+2];
            project4points[i*3+1] /= project4points[i*3+2];
            project4points[i*3+2] = 1.0;
        }
        ptpairs[i*6  ] = canonic4points[i*3  ];
        ptpairs[i*6+1] = canonic4points[i*3+1];
        ptpairs[i*6+2] = canonic4points[i*3+2];
        ptpairs[i*6+3] = project4points[i*3  ];
        ptpairs[i*6+4] = project4points[i*3+1];
        ptpairs[i*6+5] = project4points[i*3+2];
    }
    //return isdegenerate<ProjectiveTransformation::Homography, DegeneracyCheck::Level2>(ptpairs, 6);
	return isdegenerate_homography_2(ptpairs, 6);
}

bool isdegenerate(ProjectiveTransformation::WarpModel w, DegeneracyCheck::Level c, const double *x, const int m)
{
	if(w == ProjectiveTransformation::Similarity && c == DegeneracyCheck::Level0)
		return isdegenerate_similarity_0(x,m);
	else if(w == ProjectiveTransformation::Affine && c == DegeneracyCheck::Level0)
		return isdegenerate_affine_0(x,m);
	else if(w == ProjectiveTransformation::Homography && c == DegeneracyCheck::Level0)
		return isdegenerate_homography_0(x,m);
	else if(w == ProjectiveTransformation::Affine && c == DegeneracyCheck::Level1)
		return isdegenerate_affine_1(x,m);
	else if(w == ProjectiveTransformation::Homography && c == DegeneracyCheck::Level1)
		return isdegenerate_homography_1(x,m);
	else if(w == ProjectiveTransformation::Affine && c == DegeneracyCheck::Level2)
		return isdegenerate_affine_2(x,m);
	else if(w == ProjectiveTransformation::Homography && c == DegeneracyCheck::Level2)
		return isdegenerate_homography_2(x,m);
	else if(w == ProjectiveTransformation::Affine && c == DegeneracyCheck::Level3)
		return isdegenerate_affine_3(x,m);
	else if(w == ProjectiveTransformation::Homography && c == DegeneracyCheck::Level3)
		return isdegenerate_homography_3(x,m);

	return false;
}
