// ransacfithomography.cpp : Defines the entry point for the application.

#include "ransacfithomography.h"
#include "homography2d.h"
#include <iostream>
#include <algorithm>
#include <vector>

//careful::modified
#include "ransac.h"
//#define _VERBOSE_FLAG_
/*
* RANSACFITHOMOGRAPHY - fits 2D homography using RANSAC
*
* Usage:   int ransacfithomography(x1, x2, rows, npts, t, H, inliers)
*
* Inputs:
*          x1  - rows x npts set of homogeneous points.  If the data is
*                2xN it is assumed the homogeneous scale factor is 1.
*          x2  - rows x npts set of homogeneous points such that x1<->x2.
*				 x1 and x2 are 1D double arrays.
*		   rows- Dimensionality of input data. 2 if nonhomogeneous data,
*				 3 if homogeneous data.
*		   npts- Number of points.
*          t   - The distance threshold between data point and the model
*                used to decide whether a point is an inlier or not. 
*                Note that point coordinates are normalized to that their
*                mean distance from the origin is sqrt(2).  The value of
*                t should be set relative to this, say in the range 
*                0.001 - 0.01  
*
* Outputs:
*          H       - The 3x3 homography such that x2 = H*x1. (Column major)
*          inliers - An array of indices of the elements of x1, x2 that were
*                    the inliers for the best model.
*		   Function return - The number of the inliers and ransac information.
*				= -1: Must have at least 4 points to fit homography;
*				= -2: x1 and x2 must have 2 or 3 rows;
*				= -3: No least-square solution for the inliers;
*				= -4: No nondegenerate data set;
*				= -5: Ransac has no useful solution
* Reference:
*		Adapted from Peter Kovesi's matlab function code:
*		http://www.csse.uwa.edu.au/~pk/research/matlabfns/
*/
/*
template<ProjectiveTransformation::WarpModel wm, DegeneracyCheck::Level dc, const int minSizeSample>
int ransacfithomography(const double *x1, const double *x2, const int rows, const int npts,
                        const double t, double H[9], int *inliers)
					*/
int ransacfithomography(ProjectiveTransformation::WarpModel wm, DegeneracyCheck::Level dc, const int minSizeSample, 
						const double *x1, const double *x2, const int rows, const int npts, 
                        const double t, double H[9], int *inliers)
{
    if (npts < minSizeSample)
    {
#ifdef _VERBOSE_FLAG_
        std::cerr<< "Must have at least 4 points to fit homography." << std::endl;
#endif
        return RansacErrorCode::RANSAC_INSUFFICIENT_POINTS;
    }

    if (rows != 2 && rows != 3)
    {
#ifdef _VERBOSE_FLAG_
        std::cerr << "x1 and x2 must have 2 or 3 rows" << std::endl;
#endif
        return RansacErrorCode::RANSAC_WRONG_DIMENSIONS;
    }

    const int rows6 = 6;
    //double *x = new double[rows6*npts];
    std::vector<double> xvec(rows6*npts);
    double *x = &xvec[0];
    if (rows == 2)
    {
        for (int i = 0; i < npts; i++)
        {
            int irows6 = i*rows6;
            int irows = i*rows;
            x[irows6  ] = x1[irows  ];
            x[irows6+1] = x1[irows+1];
            x[irows6+2] = 1.0;
            x[irows6+3] = x2[irows  ];
            x[irows6+4] = x2[irows+1];
            x[irows6+5] = 1.0;
        }
    }
    else
    {
        for (int i = 0; i < npts; i++)
        {
            int irows6 = i*rows6;
            int irows = i*rows;
            x[irows6  ] = x1[irows  ];
            x[irows6+1] = x1[irows+1];
            x[irows6+2] = x1[irows+2];
            x[irows6+3] = x2[irows  ];
            x[irows6+4] = x2[irows+1];
            x[irows6+5] = x2[irows+2];
        }
    }

    double T1[9];
    double T2[9];

    normalize2dpts(x, rows6, npts, x, T1);
    normalize2dpts(x+rows6/2, rows6, npts, x+rows6/2, T2);

    std::vector<int> inliersvec(npts);
	int ninliers;
    ninliers = ransac(wm, dc, x, rows6, npts, minSizeSample, t, H, inliersvec, 
                   isdegenerate, homography2d, homogdist2d);        
	
    std::copy(inliersvec.begin(), inliersvec.end(), inliers);

    if (ninliers >= minSizeSample)
    {
        for (int i = 0; i < ninliers; i++)
        {
            for (int j = 0; j < rows6; j++)
            {
                x[i*rows6+j] = x[inliers[i]*rows6+j];
            }
        }
		
        int info = homography2d(wm, x, rows6, ninliers, H);
        if (info != 0)
        {
            ninliers = RansacErrorCode::RANSAC_NO_LS_SOLUTION;
            return ninliers;
        }
        else
        {
            // Denormalise
            // H = T2\H*T1;
            denormalize(H, T1, T2);

        }
        if (DegeneracyCheck::check_homography)
        {
            if ( invalidhomography(H) ){
                ninliers = RansacErrorCode::RANSAC_INVALID_HOMOGRAPHY;
                return ninliers;
            }
        } // end H check
    }
   
    return ninliers;
}

#if 0
template int ransacfithomography<ProjectiveTransformation::Similarity, DegeneracyCheck::Level0, 2>
(const double *x1, const double *x2, const int rows, const int npts, const double t, double H[9], int *inliers);

template int ransacfithomography<ProjectiveTransformation::Affine, DegeneracyCheck::Level2, 3>
(const double *x1, const double *x2, const int rows, const int npts, const double t, double H[9], int *inliers);

template int ransacfithomography<ProjectiveTransformation::Homography, DegeneracyCheck::Level0, 4>
(const double *x1, const double *x2, const int rows, const int npts, const double t, double H[9], int *inliers);

template int ransacfithomography<ProjectiveTransformation::Homography, DegeneracyCheck::Level2, 4>
(const double *x1, const double *x2, const int rows, const int npts, const double t, double H[9], int *inliers);

#endif