
#include "ObjectModel.h"
#include "Utils/Utils.h"
#include "RegionSegmentation.h"
#include <iostream>
#include "Motion/MotionDetector.h"

#define MAX_COST 10000000.0f
#define MIN_COST 0.0f

// operator overloaded to display the region attributes
ostream& operator <<(ostream& out, const RegionAttributes& att)
{
	int i, j;
	cout << "Area: " << att.m_area << endl;
	cout << "Label: " << att.m_label << endl;
	cout << "Centroid: " << att.m_centroid.x << ", " << att.m_centroid.y << endl;
	cout << "Tracking Context: [" << att.m_tracking_context.left << ", " << att.m_tracking_context.top << ", " << att.m_tracking_context.right << ", " << att.m_tracking_context.bottom << "]" << endl;
	cout << "Bounding Box: [" << att.m_bounding_box.left << ", " << att.m_bounding_box.top << ", " << att.m_bounding_box.right << ", " << att.m_bounding_box.bottom << "]" << endl;
	cout << "Region Mean: " << att.m_region_mean[0] << ", " << att.m_region_mean[1] << endl;
	cout << "Region Covariance: " << endl << att.m_region_cov[0][0] << " " << att.m_region_cov[0][1] << endl << att.m_region_cov[1][0] << " " << att.m_region_cov[1][1] << endl;
	cout << "Feature Mean: ";
	for(i=0; i<3; i++)
		cout << att.m_feature_mean[i] << " ";
	cout << endl;

	cout << "Feature Covariance: " << endl;
	for(i=0; i<3; i++) {
		for(j=0; j<3; j++)
			cout << att.m_feature_cov[i][j] << " ";
		cout << endl;
	}

	cout << "------------------------------" << endl;
	return out;
}

ObjectModel::ObjectModel(int width, int height) 
	:m_width(width), m_height(height)
{
	m_tc_threshold = 50;
	m_edge_image.Reset(width, height);
	m_no_past_frames = 100;
	m_update_weight = 0.01f;
}

void ObjectModel::CalcObjectAttributes(const ImgInt& labels, RegionAttributes& rgn_att)
{
	int x, y;
	int width = labels.Width();
	int height = labels.Height();

	// calculate the mean
	ImgInt::ConstIterator p_labels = labels.Begin();
	int x_mean=0, y_mean=0, area=0;
	rgn_att.m_pixel_list.clear();

	for(y=0; y<height; y++) {
		for(x=0; x<width; x++, p_labels++) {
			if(*p_labels > 0) {
				area++;
				x_mean += x;
				y_mean += y;
				rgn_att.m_pixel_list.push_back(Point(x,y));
			}
		}
	}
	x_mean /= area; y_mean /= area;
	
	rgn_att.vel_x = 0; rgn_att.vel_y = 0;
	rgn_att.m_centroid.x = x_mean;rgn_att.m_centroid.y = y_mean;
	ComputeSpatialModel(rgn_att.m_pixel_list, rgn_att, x_mean, y_mean, area);
}

void ObjectModel::CalcObjectAttributes(const ImgInt& labels)
{
	CalcObjectAttributes(labels, m_object_att);
}

bool ObjectModel::UpdateObjectAttributes(const ImgBinary& mask, RegionAttributes& rgn_att, int& dx, int& dy)
{
	int x, y;
	Rect tc = rgn_att.m_tracking_context;
	rgn_att.m_pixel_list.clear();
	ImgBinary::ConstIterator p_mask = mask.Begin();

	int height = mask.Height();
	int width = mask.Width();

	int x_mean=0, y_mean=0, area=0;
	for(y=0; y<height; y++) {
		for(x=0; x<width; x++, p_mask++) {
			if(*p_mask) {
				area++;
				x_mean += x;
				y_mean += y;
				rgn_att.m_pixel_list.push_back(Point(x,y));				
			}
		}
	}

	if(area == 0) {
		printf("object has vanished\n");
		return false;
	}
	x_mean /= area; y_mean /= area;

	// Calculate dx and dy
	dx = x_mean - rgn_att.m_centroid.x;
	dy = y_mean - rgn_att.m_centroid.y;

	int size = rgn_att.m_pixel_list.size();
	if(size < 10) {
		printf("Object strength is too less.. Tracking might fail.\n");
		return false;
	}

	// Compute the spatial model
	ComputeSpatialModel(rgn_att.m_pixel_list, rgn_att, x_mean, y_mean, area);
	
	/*
	if(dx > 20 || dy > 20)
		printf("Object moving very fast at [%d %d] !!!\n", dx, dy);
	*/

	return true;
}

bool ObjectModel::UpdateObjectAttributes(const ImgBinary& mask, int& dx, int& dy)
{
	return UpdateObjectAttributes(mask, m_object_att, dx, dy);
}

void ObjectModel::ComputeSpatialModel(const list<Point>& pixel_list, RegionAttributes& rgn_att)
{
	int size = pixel_list.size();
	int area = 0, x_mean = 0, y_mean = 0;
	for(list<Point>::const_iterator list_iter = pixel_list.begin(); list_iter != pixel_list.end(); list_iter++) {
		area++;
		x_mean += (*list_iter).x;
		y_mean += (*list_iter).y;
	}
	x_mean /= area;
	y_mean /= area;
	ComputeSpatialModel(pixel_list, rgn_att, x_mean, y_mean, area);
}

void ObjectModel::ComputeSpatialModel(const list<Point>& pixel_list, RegionAttributes& rgn_att, int x_mean, int y_mean, int area)
{
	int min_x, min_y, max_x, max_y;

	// Initialize the region covariance
	double region_cov[2][2];
	region_cov[0][0] = 0; region_cov[0][1] = 0; region_cov[1][0] = 0; region_cov[1][1] = 0;

	// initialize tracking context
	min_x = m_width+1; max_x = 0;
	min_y = m_height+1; max_y = 0;

	// Calculate the region covariance
	for(list<Point>::const_iterator list_iter = pixel_list.begin(); list_iter != pixel_list.end(); list_iter++) {
		const Point &pt = *list_iter;
		region_cov[0][0] += ((pt.x - x_mean) * (pt.x - x_mean));	// x,x
		region_cov[1][0] += ((pt.x - x_mean) * (pt.y - y_mean));	// x,y
		region_cov[1][1] += ((pt.y - y_mean) * (pt.y - y_mean));	// y,y
		if(pt.x < min_x)	min_x = pt.x;		// tracking context is formed using min and max (x,y)'s
		if(pt.y < min_y)	min_y = pt.y;
		if(pt.x > max_x)	max_x = pt.x;
		if(pt.y > max_y)	max_y = pt.y;
		
	}
	region_cov[0][0] /= area;
	region_cov[1][0] /= area;
	region_cov[1][1] /= area;
	region_cov[0][1] = region_cov[1][0];

	// tracking context
	int left = (min_x-m_tc_threshold < 0) ? 0 : min_x-m_tc_threshold;
	int right = (max_x+m_tc_threshold >= m_width) ? m_width-1 : max_x+m_tc_threshold;
	int top = (min_y-m_tc_threshold < 0) ? 0 : min_y-m_tc_threshold;
	int bottom = (max_y+m_tc_threshold >= m_height) ? m_height-1 : max_y+m_tc_threshold;

	// set the values in the structure
	rgn_att.vel_x = x_mean - rgn_att.m_centroid.x;
	rgn_att.vel_y = y_mean - rgn_att.m_centroid.y;
	rgn_att.m_area = area;
	rgn_att.m_centroid = Point(x_mean, y_mean);
	rgn_att.m_tracking_context = Rect(left, top, right, bottom);
	rgn_att.m_bounding_box = Rect(min_x, min_y, max_x, max_y);
	rgn_att.SetRegionMean((double)x_mean, (double)y_mean);
	rgn_att.SetRegionCovariance(region_cov);
	Utils::Inverse2x2( rgn_att.m_region_cov, rgn_att.m_inv_region_cov);

}

void ObjectModel::UpdateSpatialModel(RegionAttributes& object_att, vector<RegionAttributes>& rgn_att, const ImgBinary& mask, int dx, int dy, bool check_consistency)
{
	int i;
	int num_regions = rgn_att.size();
	
	// erase all pixels that are not inside the object contour
	for(i=0; i<num_regions; i++) {
		for(list<Point>::iterator list_iter = rgn_att[i].m_pixel_list.begin(); list_iter != rgn_att[i].m_pixel_list.end(); list_iter++) {
			Point pt = *list_iter;
			if(!mask(pt.x, pt.y)) {
				list_iter = rgn_att[i].m_pixel_list.erase(list_iter);
				list_iter--;
			}
		}		
	}

	if(check_consistency) {
		// check the consistency of the pixels with its previous spatial model
		double val = 0.0f;
		double input_vec[2];
		for(i=0; i<num_regions;i++) {
			for(list<Point>::iterator list_iter = rgn_att[i].m_pixel_list.begin(); list_iter != rgn_att[i].m_pixel_list.end(); list_iter++) {
				input_vec[0] = (*list_iter).x; input_vec[1] = (*list_iter).y;
				Utils::MahDistance2x2(input_vec, rgn_att[i].m_region_mean, rgn_att[i].m_inv_region_cov, val);
				if(val < 0.0001f) {  // used for elmo
				//if(val < 0.000000005f) {  // used for monkey
					list_iter = rgn_att[i].m_pixel_list.erase(list_iter);
					list_iter--;
				}
			}
		}
	}

	for(i=0; i<num_regions; i++) {
		int num_pixels = rgn_att[i].m_pixel_list.size();
		if(num_pixels < 10) {
			//printf("Region is too small to update the spatial model. Assigning the spatial model of the object to the region.\n");
			CopySpatialModel(object_att, rgn_att[i]);
		}
		else if(num_pixels/rgn_att[i].m_area > 5) {
			//printf("Region %d has either increased or decreased by a huge factor. Assigning the spatial model of the object to the region.\n", i);
			CopySpatialModel(object_att, rgn_att[i]);
		}
		else {
			Point old_centroid = rgn_att[i].m_centroid;
			ComputeSpatialModel(rgn_att[i].m_pixel_list, rgn_att[i]);
			int r_dx = rgn_att[i].m_centroid.x - old_centroid.x;
			int r_dy = rgn_att[i].m_centroid.y - old_centroid.y;
			/*
			if(abs(r_dx - dx) > 30 || abs(r_dy - dy) > 30)
				printf("Region movement is not in accordance with the object movement. Something might be going wrong.\n");
			*/
		}
	}
}

void ObjectModel::CopySpatialModel(RegionAttributes& src, RegionAttributes& dest)
{
	dest.m_area = src.m_area;
	dest.m_centroid = src.m_centroid;
	dest.SetRegionMean(src.m_region_mean[0], src.m_region_mean[1]);
	dest.SetRegionCovariance(src.m_region_cov);
	dest.m_tracking_context = src.m_tracking_context;
	dest.m_bounding_box = src.m_bounding_box;
	Utils::Inverse2x2( dest.m_region_cov, dest.m_inv_region_cov);
}

void ObjectModel::ComputeAppearanceModel(const ImgBgr& input_image, const list<Point>& pixel_list, RegionAttributes& rgn_att)
{
	int i, j;
	vector<float> mean;
	mean.push_back(0); mean.push_back(0); mean.push_back(0);
	for(list<Point>::const_iterator list_iter = pixel_list.begin(); list_iter != pixel_list.end(); list_iter++) {
		Point pt = *list_iter;
		mean[0] += input_image(pt.x, pt.y).r;
		mean[1] += input_image(pt.x, pt.y).g;
		mean[2] += input_image(pt.x, pt.y).b;
	}
	int size = pixel_list.size();
	mean[0] /= size; mean[1] /= size; mean[2] /= size;

	double feature_cov[3][3];
	for(i=0; i<3; i++)
		for(j=0; j<3; j++)
			feature_cov[i][j] = 0;
	
	for(list_iter = pixel_list.begin(); list_iter != pixel_list.end(); list_iter++) {
		Point pt = *list_iter;
			feature_cov[0][0] += ((input_image(pt.x, pt.y).r - mean[0]) * (input_image(pt.x, pt.y).r - mean[0]));   // r,r
			feature_cov[1][0] += ((input_image(pt.x, pt.y).r - mean[0]) * (input_image(pt.x, pt.y).g - mean[1]));	// r,g
			feature_cov[2][0] += ((input_image(pt.x, pt.y).r - mean[0]) * (input_image(pt.x, pt.y).b - mean[2]));	// r,b
			feature_cov[1][1] += ((input_image(pt.x, pt.y).g - mean[1]) * (input_image(pt.x, pt.y).g - mean[1]));	// g,g
			feature_cov[2][1] += ((input_image(pt.x, pt.y).g - mean[1]) * (input_image(pt.x, pt.y).b - mean[2]));	// g,b
			feature_cov[2][2] += ((input_image(pt.x, pt.y).b - mean[2]) * (input_image(pt.x, pt.y).b - mean[2]));	// b,b
	}
	feature_cov[0][1] = feature_cov[1][0];
	feature_cov[0][2] = feature_cov[2][0];
	feature_cov[1][2] = feature_cov[2][1];
	
	for(i=0; i<3; i++)
		for(j=0; j<3; j++)
			feature_cov[i][j] /= size;

	rgn_att.SetFeatureMean(mean[0], mean[1], mean[2]);
	rgn_att.SetFeatureCovariance(feature_cov);
	rgn_att.m_isValid = Utils::Inverse3x3(feature_cov, rgn_att.m_inv_feature_cov);
}

void ObjectModel::UpdateObjectHistory(int frame_no, const vector<Point>& contour_pts, const RegionAttributes& object_att, bool isObjectOccluded, vector<Shape> *contour_history)
{
	int i;
	//Store Contour Points in history
	vector<Point> contourOrigShifted;
	for(i=0;i<contour_pts.size();i++)
		contourOrigShifted.push_back(Point(contour_pts[i].x-object_att.m_centroid.x,-1*(object_att.m_centroid.y-contour_pts[i].y)));

	Shape objhist;
	objhist.contour_points = contourOrigShifted;
	objhist.centroid = object_att.m_centroid;
	objhist.size = object_att.m_pixel_list.size();
	objhist.IsObjectOccluded = isObjectOccluded;
	(*contour_history).push_back(objhist);
}

void ObjectModel::UpdateObjectHistory(int frame_no, bool isObjectOccluded)
{
	UpdateObjectHistory(frame_no, m_contour_points, m_object_att, isObjectOccluded, &m_contour_history);
}

void ObjectModel::GenerateObjectModel(const ImgBgr& input_image)
{
	int i;
	int no_fg_regions = m_fg_rgn_att.size();
	int no_bg_regions = m_bkg_rgn_att.size();

	// Generate foreground model for all the regions
	for(i=0; i<no_fg_regions; i++) {
		m_fg_rgn_att[i].m_label = i+1;
		ComputeSpatialModel(m_fg_rgn_att[i].m_pixel_list, m_fg_rgn_att[i]);
		ComputeAppearanceModel(input_image, m_fg_rgn_att[i].m_pixel_list, m_fg_rgn_att[i]);
	}
	// Generate background model for all the regions
	for(i=0; i<no_bg_regions; i++) {
		m_bkg_rgn_att[i].m_label = i+1;
		ComputeSpatialModel(m_bkg_rgn_att[i].m_pixel_list, m_bkg_rgn_att[i]);
		ComputeAppearanceModel(input_image, m_bkg_rgn_att[i].m_pixel_list, m_bkg_rgn_att[i]);
		m_bkg_rgn_att[i].m_is_occluded = false;
	}
	// Generate the spatial model of the object
	m_object_att.m_pixel_list.clear();
	m_object_att.m_area = 0;
	for(i=0; i<no_fg_regions; i++) {
		m_object_att.m_area += m_fg_rgn_att[i].m_area;
		for(list<Point>::iterator list_iter = m_fg_rgn_att[i].m_pixel_list.begin(); list_iter != m_fg_rgn_att[i].m_pixel_list.end(); list_iter++) {
			m_object_att.m_pixel_list.push_back(*list_iter);
		}
	}
	ComputeSpatialModel(m_object_att.m_pixel_list, m_object_att);

	// Copy this fg model into initial foreground model
	m_fg_initial_model = m_fg_rgn_att;
	m_fg_rgn_history.push_back(m_fg_initial_model);

	// Copy this bg model into initial background model
	m_bg_initial_model = m_bkg_rgn_att;
	m_bg_rgn_history.push_back(m_bg_initial_model);
}

/*
 Update edge.. This is done before tracking the current frame. Unlike updates of color n spatial model, 
 updating edges are done independent of tracking. Canny edge is performed on every frame and this result is 
 combined with the strength map during the strength map computation in ObjectTracker.cpp
*/
void ObjectModel::UpdateEdge(const ImgBgr& input_image) 
{
	ImgGray image_gray;
	Convert(input_image, &image_gray);
	Canny(image_gray, &m_edge_image, 3.0f);
	Dilate3x3(m_edge_image, &m_edge_image);
}

void ObjectModel::UpdateFgSpatialModel(const ImgBinary& mask, int dx, int dy, bool check_consistency)
{
	UpdateSpatialModel(m_object_att, m_fg_rgn_att, mask, dx, dy, check_consistency);
}

void ObjectModel::UpdateAppearanceModel(const ImgBgr& current_frame, vector<RegionAttributes>& rgn_att, vector<RegionAttributes>& initial_model, deque< vector< RegionAttributes > > *rgn_history)
{
	int i, p, q;
	
	double tau, T, weight;
	double mean[3], cov[3][3], inv_cov[3][3]; 
	double normalization_val;
	list<Point>::iterator list_iter;
	double alpha, beta1, beta2;
	double feature_vec[3], dist;

	try {
	for(i=0; i<rgn_att.size(); i++) {
		// If the fragment is too small, continue
		if(rgn_att[i].m_pixel_list.size() < 100)
			continue;

		// Traverse all the pixel list in the fragment. If the pixel is an edge pixel, remove it from the list.
		list<Point> tmp_list = rgn_att[i].m_pixel_list;
		for(list_iter = tmp_list.begin(); list_iter != tmp_list.end(); list_iter++) {
			Point pt = *list_iter;
			if(m_edge_image(pt.x, pt.y)) {
				list_iter = tmp_list.erase(list_iter);
				list_iter--;
			}
		}

		// Calculate the current mean and covariance of the appearance based on the 'revised' pixel list in the fragment
		RegionAttributes tmp_rgn_att;
		ComputeAppearanceModel(current_frame, tmp_list, tmp_rgn_att);

		// Calculate the past statistics (0-t) as given in eqn (21, 22).
		T = (*rgn_history).size();
		mean[0] = 0; mean[1] = 0; mean[2] = 0;
		for(p=0; p<3; p++)
			for(q=0; q<3; q++)
				cov[p][q] = 0;
		normalization_val = 0;
		// statistics of frames: 0 to t-1
		for(tau=0; tau<T; tau++) {		
			weight = exp(-m_update_weight * (T-tau));
			mean[0] += weight * (*rgn_history)[tau][i].m_feature_mean[0];
			mean[1] += weight * (*rgn_history)[tau][i].m_feature_mean[1];
			mean[2] += weight * (*rgn_history)[tau][i].m_feature_mean[2];
			for(p=0; p<3; p++)
				for(q=0; q<3; q++)
					cov[p][q] += weight * (*rgn_history)[tau][i].m_feature_cov[p][q];
			normalization_val += weight;
		}
		// add the statistics of the current frame
		mean[0] += tmp_rgn_att.m_feature_mean[0];
		mean[1] += tmp_rgn_att.m_feature_mean[1];
		mean[2] += tmp_rgn_att.m_feature_mean[2];
		for(p=0; p<3; p++)
			for(q=0; q<3; q++)
				cov[p][q] += tmp_rgn_att.m_feature_cov[p][q];
		normalization_val += 1.0f;

		// Normalize the mean and covariance
		mean[0] /= normalization_val; mean[1] /= normalization_val; mean[2] /= normalization_val;
		for(p=0; p<3; p++)
			for(q=0; q<3; q++)
				cov[p][q] /= normalization_val;
		Utils::Inverse3x3(cov, inv_cov);

		// Calculate the weight - alpha
		beta1 = 0; beta2 = 0;
		for(list_iter = tmp_list.begin(); list_iter != tmp_list.end(); list_iter++) {
			Point pt = *list_iter;
			feature_vec[0] = current_frame(pt.x, pt.y).r; feature_vec[1] = current_frame(pt.x, pt.y).g; feature_vec[2] = current_frame(pt.x, pt.y).b;
			Utils::MahDistance3x3(feature_vec, initial_model[i].m_feature_mean, initial_model[i].m_inv_feature_cov, dist);
			beta1 += dist;
			Utils::MahDistance3x3(feature_vec, mean, inv_cov, dist);
			beta2 += dist;
		}
		alpha = beta1 / (beta1 + beta2);

		// Calculate the new mean and covariance as given in eqn (19, 20)
		rgn_att[i].m_feature_mean[0] = (alpha * mean[0]) + ((1 - alpha) * initial_model[i].m_feature_mean[0]);		
		rgn_att[i].m_feature_mean[1] = (alpha * mean[1]) + ((1 - alpha) * initial_model[i].m_feature_mean[1]);		
		rgn_att[i].m_feature_mean[2] = (alpha * mean[2]) + ((1 - alpha) * initial_model[i].m_feature_mean[2]);
		for(p=0; p<3; p++)
			for(q=0; q<3; q++)
				rgn_att[i].m_feature_cov[p][q] = (alpha * cov[p][q]) + ((1 - alpha) * initial_model[i].m_feature_cov[p][q]);	
		rgn_att[i].m_isValid = Utils::Inverse3x3(rgn_att[i].m_feature_cov, rgn_att[i].m_inv_feature_cov);

		rgn_att[i].m_pixel_list = tmp_list;
	}

	// update the object history. If there are more than N elements, remove the first element and insert the latest one
	if((*rgn_history).size() > m_no_past_frames) {
		(*rgn_history).pop_front();
		(*rgn_history).push_back(rgn_att);
	}
	else
		(*rgn_history).push_back(rgn_att);
	}
	catch(...) {
		cout << "Crashed in UpdateAppearance()" << endl;
	}
}

void ObjectModel::UpdateFgAppearanceModel(const ImgBgr& current_frame)
{
	m_update_weight = 10.01f;
	UpdateAppearanceModel(current_frame, m_fg_rgn_att, m_fg_initial_model, &m_fg_rgn_history);
}

void ObjectModel::UpdateBgSpatialModel(const ImgBinary& mask)
{
	int i;
	int num_regions = m_bkg_rgn_att.size();

	// Remove all the pixels in the foreground regions
	for(i=0; i<num_regions; i++) {
		for(list<Point>::iterator list_iter = m_bkg_rgn_att[i].m_pixel_list.begin(); list_iter != m_bkg_rgn_att[i].m_pixel_list.end(); list_iter++) {
			if(mask((*list_iter).x, (*list_iter).y)) {
				list_iter = m_bkg_rgn_att[i].m_pixel_list.erase(list_iter);
				list_iter--;
			}
		}
	}

	// Update the spatial model
	for(i=0; i<num_regions; i++) {
		int num_pixels = m_bkg_rgn_att[i].m_pixel_list.size();
		if(num_pixels > 20)
			ComputeSpatialModel(m_bkg_rgn_att[i].m_pixel_list, m_bkg_rgn_att[i]);
	}

}

void ObjectModel::UpdateBgAppearanceModel(const ImgBgr& current_frame)
{
	m_update_weight = 10.01f;
	UpdateAppearanceModel(current_frame, m_bkg_rgn_att, m_bg_initial_model, &m_bg_rgn_history);
}

void ObjectModel::UpdateNewModes(const ImgBgr& current_frame, const ImgBgr& previous_frame, const ImgFloat& strength_image, const ImgBinary& object_mask)
{
	int x, y;

	// Generate the mask from the strength image. Pixels whose value are close to 0 are set as true
	ImgBinary mask(m_width, m_height);
	for(y=0; y<m_height; y++) {
		for(x=0; x<m_width; x++) {
			mask(x,y) = (strength_image(x,y) <= 0.000000000000001f && strength_image(x,y) >= -0.0f) ? true : false;
		}
	}
	static Figure new_mask_wnd("Mask for updating new modes");
	new_mask_wnd.Draw(mask);

	/*  
		Check if white regions are obtained in the mask for self-occluding cases.
		If it happens, this will work. The next step would be to check if 
		there are 10 pixels in the white region that are 4-connected with each other 
		and 4-connected with the dtected object. If such a region appears for 3 contiguous frames, 
		then that region's appearance can be segmented to check if it has a single mode. If it has a single mode,
		it can be updated in the fg model.

		NOTE: During that phase of 3 frames, those pixels should not be considered for updating the existing fg and bg mode.
		Since the strength image is 0, and pixels outside the contour are removed before adding them to the fg, I hope it is
		not added to fg model. What abt bg model ? Is it being added ? If so, that has to be removed using this 
		mask generated from strength image.
	*/

	// Get the connected components of the regions in the mask
	vector< ConnectedComponentProperties<ImgBinary::Pixel> > cc;
	ImgInt labels(m_width, m_height);
	ConnectedComponents4(mask, &labels, &cc);

	// Remove all regions whose area is < 50
	vector< ConnectedComponentProperties<ImgBinary::Pixel> >::iterator iter;
	for(iter = cc.begin(); iter != cc.end(); iter++) {
		if((*iter).npixels <= 100 || (*iter).npixels >= 500) {
			FloodFill4(labels, (*iter).pixel.x, (*iter).pixel.y, 0, &labels);
			FloodFill4(mask, (*iter).pixel.x, (*iter).pixel.y, 0, &mask);
			iter = cc.erase(iter);
			iter--;
		}
	}

	if(cc.size() == 1)
		return;

	// ----------------------------------------------
	ImgBgr out;
	Utils::RandomColor(labels, &out);
	static Figure cc_wnd("Connected Components output");
	cc_wnd.Draw(out);
	// ----------------------------------------------
	
	int num_regions = cc.size();
	vector<int> new_rgns;
	for(int i=1; i<num_regions; i++) {
		Rect bb = cc[i].bounding_rect;
		// increase the rect area and check if any strength is there in this area
		bb.left -= 5; bb.right += 5; bb.top -= 5; bb.bottom += 5;
		if(bb.left < 0) bb.left = 0;
		if(bb.right >= m_width) bb.right = m_width-1;
		if(bb.top < 0) bb.top = 0;
		if(bb.bottom >= m_height) bb.bottom = m_height-1;
		// Count the no of strength (fg) pixels in this rect area.
		int cnt = 0;
		for(y=bb.top; y<=bb.bottom; y++)
			for(x=bb.left; x<=bb.right; x++)
				//if(strength_image(x,y) > 0) cnt++;
				if(object_mask(x,y)) cnt++;
		// if more than 20 strength pixels are available, then declare this region as adjacent to fg else add it to bg
		if(cnt >= 40)
			new_rgns.push_back(i);
	}

	/*
	ImgBgr disp_image = current_frame;
	for(vector<int>::iterator it = new_rgns.begin(); it != new_rgns.end(); it++) {
		Rect bb = cc[*it].bounding_rect;
		// increase the rect area
		bb.left -= 20; bb.right += 20; bb.top -= 20; bb.bottom += 20;
		if(bb.left < 0) bb.left = 0;
		if(bb.right >= m_width) bb.right = m_width-1;
		if(bb.top < 0) bb.top = 0;
		if(bb.bottom >= m_height) bb.bottom = m_height-1;
		// compute average motion of the new region
		float rgn_du = 0.0f, rgn_dv = 0.0f;
		MotionDetector::ComputeJointKltMotion(previous_frame, current_frame, mask, bb, rgn_du, rgn_dv, disp_image, 3, 8, 0);
		// compute average motion of the fg region near the new region
		float fg_du = 0.0f, fg_dv = 0.0f;
		ImgBinary tmp_mask(m_width, m_height);
		Set(&tmp_mask, 0);
		for(y=bb.top; y<=bb.bottom; y++)
			for(x=bb.left; x<=bb.right; x++)
				if(strength_image(x,y) > 0) tmp_mask(x,y) = true;
		MotionDetector::ComputeJointKltMotion(previous_frame, current_frame, tmp_mask, bb, fg_du, fg_dv, disp_image, 3, 8, 1);
		// if the two motions are similar, then retain the region. otherwise delete it
		cout << rgn_du << ", " << rgn_dv << " | " << fg_du << ", " << fg_dv << endl;
		if(abs(fg_du - rgn_du) > 3 || abs(fg_dv - rgn_dv) > 3) {
			it = new_rgns.erase(it);
			it--;
		}

	}
	static Figure rgn_mot_wnd("Motion of new regions");
	rgn_mot_wnd.Draw(disp_image);
	*/

	/*
		segment the new regions using a high sigma value and create the appearance model
		NOTE: Due to my languidness, I've assumed the new regions to be unimodal. 
		Ideally, this has to go through the segmentation process !
	*/
	if(new_rgns.size()) {
		cout << new_rgns.size() << " region(s) found" << endl;
		for(vector<int>::iterator it = new_rgns.begin(); it != new_rgns.end(); it++) {
			AddNewMode(cc[*it].bounding_rect, mask, current_frame, m_fg_rgn_att, m_fg_initial_model);
			cout << "Region added..." << endl;
		}
		getchar();
	}
}

void ObjectModel::AddNewMode(const Rect& bb, const ImgBinary& mask, const ImgBgr& current_frame, vector<RegionAttributes>& model, vector<RegionAttributes>& initial_model)
{
	int x, y;
	list<Point> point_list;
	for(y=bb.top; y<=bb.bottom; y++)
		for(x=bb.left; x<=bb.right; x++)
			if(mask(x,y)) point_list.push_back(Point(x,y));

	RegionAttributes rgn_att;
	ComputeAppearanceModel(current_frame, point_list, rgn_att);
	ComputeSpatialModel(point_list, rgn_att);
	rgn_att.m_pixel_list = point_list;
	rgn_att.m_area = point_list.size();
	model.push_back(rgn_att);
	initial_model.push_back(rgn_att);
	

	cout << rgn_att << endl;
}