
#include "ObjectTracker.h"
#include "utils/ScanfillSkeleton.h"
#include <stack>
#include <iostream>
#include "utils/Utils.h"
#include "Motion/MotionDetector.h"

#define MAX_BG_STRENGTH 1
#define EDGE_STRENGTH 1
//#define _OCCLUSSION_

ObjectTracker::ObjectTracker(int width, int height)
	:m_width(width), m_height(height), m_contour_extractor(width, height), m_occlusion_handler(width, height), m_object_model(width, height)
{
}

ObjectTracker::~ObjectTracker()
{
	m_segmentation_handler.FreeMemory();
}
int ObjectTracker::Initialize(const ImgBgr &input_image, vector<Point>& points, int obj_id, double seg_sigma, int num_features, ImgInt *const fg_label, ImgInt *const bg_label)
{
	m_frame_count = -1;
	m_obj_id = obj_id;

	// Initialize all the scratch images and parameters to default values
	m_distance_threshold = 50;
	
	m_num_features = num_features;
	m_object_model.m_contour_points = points;
	m_nlevels = 3;
	m_mindist = 8;
	m_seg_sigma = seg_sigma;

	m_use_motion = false;
	m_use_region_consistency = false;
	m_use_edges = false;
	m_contour_method = WALL_FOLLOW;

	m_strength_image.Reset(m_width, m_height);
	m_fg_strength.Reset(m_width, m_height);
	m_bg_strength.Reset(m_width, m_height);
	m_object_mask.Reset(m_width, m_height);
	FillPolygon(&m_object_mask, m_object_model.m_contour_points);

	srand(clock());
	//m_contour_color = ImgBgr::Pixel(((rand()%10) * 12342223) % 0xFFFFFF, Bgr::BLEPO_BGR_XBGR);
	m_contour_color = Bgr(0,255,0);
	
	// Generate the mask
	ImgBinary input_mask;
	input_mask.Reset(m_width, m_height);
	Set(&input_mask, 0);
	FillPolygon(&input_mask, points);
	
	// Segment the image and find the foreground and background regions
	m_segmentation_handler.DoSegmentation(input_image, true, input_mask, m_seg_sigma, &m_object_model.m_fg_rgn_att, &m_object_model.m_bkg_rgn_att);

	// Generate the spatial and appearance model of all regions and the spatial model of the object
	m_object_model.GenerateObjectModel(input_image);

	m_prev_frame = input_image;
	return 1;
}


/*
  Computes the strength map by calculating the log ratio of the foreground and background probabilities.
  Update m_fg_strength, m_bg_strength and m_strength_image. It also updates the pixel list of each of the fragments.
*/
void ObjectTracker::_ComputeStrengthMap(const ImgBgr& current_frame)
{
	int x, y, i;

	double fg_strength, bg_strength, spatial_strength, max, bg_max;
	double total_fg_strength = 0.0f, total_bg_strength = 0.0f;
	int rgn_index, bg_rgn_index;
	double r_eps;
	//r_eps = 0.00000000000001f;		// monkey
	//r_eps = 0.0000001f;		// fish
	r_eps = 0.000000000000000000000000000000000000000000001f;  // elmo, pets3
	//r_eps = 0.0000000000000000000000000000000000000000001f;  // elmo self occ
	// r_eps = 0.0000000000000000000000000000001f;  // lab
	vector<int> fg_rgn_index;
	double feature_vec[3];
	double spatial_vec[2];
	double feature_strength;
	double t_spatial_strength;
	double obj_spatial_strength;

	try {
	for(y=0; y<m_height; y++) {	
		for(x=0; x<m_width; x++) {
			fg_strength = 0.0f; bg_strength = 0.0f, spatial_strength = 0.0f; max = -1; bg_max = -1; rgn_index = -1; bg_rgn_index = -1; feature_strength = 0.0f; t_spatial_strength = 0.0f;
			fg_rgn_index.clear();
			for(i=0; i<m_object_model.m_fg_rgn_att.size(); i++) {
				if(!m_object_model.m_fg_rgn_att[i].m_isValid)
					continue;

				feature_vec[0] = current_frame(x,y).r; 	feature_vec[1] = current_frame(x,y).g;  feature_vec[2] = current_frame(x,y).b;
				Utils::MahDistance3x3(feature_vec, m_object_model.m_fg_rgn_att[i].m_feature_mean, m_object_model.m_fg_rgn_att[i].m_inv_feature_cov, feature_strength);

				// Calculate the spatial weights
				spatial_vec[0] = x; spatial_vec[1] = y;
				Utils::MahDistance2x2(spatial_vec, m_object_model.m_fg_rgn_att[i].m_region_mean, m_object_model.m_fg_rgn_att[i].m_inv_region_cov, t_spatial_strength);

				// Find the foreground strength
				double temp_strength = feature_strength * t_spatial_strength;
				fg_strength += temp_strength; // * 10000.0f;
	
				// Calculate the max strength contributed by the region within the tracking context
				Rect& rect = m_object_model.m_fg_rgn_att[i].m_tracking_context;
				if(temp_strength > max && x>rect.left && x<rect.right && y>rect.top && y<rect.bottom) {
					max = temp_strength;
					rgn_index = i;
				}

				// If the current point is in the tracking context of a foreground region, push the foreground index into a vector.. 
				// Used while computing the background strength
				if(x>rect.left && x<rect.right && y>rect.top && y<rect.bottom) 
					fg_rgn_index.push_back(i);
			}

			// Add the edge strength
			if(m_use_edges) {
				if(m_object_model.m_edge_image(x,y)) {
					// Find the object spatial strength
					spatial_vec[0] = x; spatial_vec[1] = y;
					Utils::MahDistance2x2(spatial_vec, m_object_model.m_object_att.m_region_mean, m_object_model.m_object_att.m_inv_region_cov, obj_spatial_strength);
					// edge strength = edge_indicator * object spatial strength
					double edge_strength = EDGE_STRENGTH * obj_spatial_strength;
					fg_strength += edge_strength;
				}
			}

			m_fg_strength(x,y) = fg_strength;
			total_fg_strength += fg_strength;

			// if the pixel was in the TC of atleast one of the fg regions, then push the point into the corresponding region
			// which contributed the max strength (obtained from rgn_index)
			if(max != -1)
				m_object_model.m_fg_rgn_att[rgn_index].m_pixel_list.push_front(Point(x,y));

			// Calculate the background strength
			if(fg_rgn_index.size() == 0)			// if the current point is far from all the foreground regions, set a high strength
				bg_strength = MAX_BG_STRENGTH;
			else {
				for(i=0; i<m_object_model.m_bkg_rgn_att.size(); i++) {					// for each bg region
					if(m_object_model.m_bkg_rgn_att[i].m_is_occluded)			// if the region is occluded, do not consider it for strength
						continue;
					if(!m_object_model.m_bkg_rgn_att[i].m_isValid)
						continue;

					feature_vec[0] = current_frame(x,y).r;
					feature_vec[1] = current_frame(x,y).g;
					feature_vec[2] = current_frame(x,y).b;
					Utils::MahDistance3x3(feature_vec, m_object_model.m_bkg_rgn_att[i].m_feature_mean, m_object_model.m_bkg_rgn_att[i].m_inv_feature_cov, feature_strength);

					// Calculate the spatial weights
					spatial_vec[0] = x; spatial_vec[1] = y;
					Utils::MahDistance2x2(spatial_vec, m_object_model.m_bkg_rgn_att[i].m_region_mean, m_object_model.m_bkg_rgn_att[i].m_inv_region_cov, t_spatial_strength);
		
					// Calculate the bg strength
					double temp_strength = feature_strength * t_spatial_strength;
					bg_strength += temp_strength;

					// Calculate the max strength contributed by the region within the tracking context
					if(temp_strength > bg_max) {
						bg_max = temp_strength;
						bg_rgn_index = i;
					}
				}
				if(bg_max != -1) 
					m_object_model.m_bkg_rgn_att[bg_rgn_index].m_pixel_list.push_front(Point(x,y));
			}
			m_bg_strength(x,y) = bg_strength;
			total_bg_strength += bg_strength;
			
			// Calculate the log ratio
			m_strength_image(x,y) = log(blepo_ex::Max(fg_strength,r_eps )/blepo_ex::Max(bg_strength,r_eps));
			if(m_strength_image(x,y) < 0 && rgn_index != -1)
				m_object_model.m_fg_rgn_att[rgn_index].m_pixel_list.pop_front();
			if(m_strength_image(x,y) > 0 && bg_rgn_index != -1)
				m_object_model.m_bkg_rgn_att[bg_rgn_index].m_pixel_list.pop_front();

		}  // x
	}  // y

	 static Figure final_strength_wnd("**Final Strength", m_width+10, 0);
	 final_strength_wnd.Draw(m_strength_image);
	}
	catch(...) {
		cout << "Crashed in _ComputeStrengthMap" << endl;
	}

}

int ObjectTracker::TrackNextFrame(ImgBgr &current_frame, int frame_no)
{
	int i;
	m_frame_count++;
	int width = current_frame.Width();
	int height = current_frame.Height();
	m_frame_no = frame_no;

	// Compute the motion vectors and update the spatial model of each fg region.
	_ComputeMotion(current_frame);

	// clear the pixels lists of all regions
	for(i=0; i<m_object_model.m_fg_rgn_att.size(); i++)
		m_object_model.m_fg_rgn_att[i].m_pixel_list.clear();
	for(i=0; i<m_object_model.m_bkg_rgn_att.size(); i++)
		m_object_model.m_bkg_rgn_att[i].m_pixel_list.clear();

	// compute the log ratio of the fg and bg strength
	_ComputeStrengthMap(current_frame);

	// Generate a mask based on the previous contour points. 
	int dx = 0, dy = 0;
	if(m_contour_method != FAST_LEVEL_SETS) {
		Set(&m_object_mask, 0);
		FillPolygon(&m_object_mask, m_object_model.m_contour_points);
	}

	// Find the contour
	if(m_contour_method == WALL_FOLLOW)
		m_contour_extractor.WallFollow(m_strength_image, m_object_model);
	else if(m_contour_method == LEVEL_SETS)
		m_contour_extractor.DoLevelSets(current_frame, m_strength_image, m_object_mask, m_object_model, &m_object_model.m_contour_points);
	else
		m_contour_extractor.FastLevelSets(current_frame, m_strength_image, m_object_mask, &m_object_model.m_contour_points);

	if(m_object_model.m_contour_points.size() == 1) {
		cout << "Something weird has happened during wall follow. Do not update anything" << endl;
		m_object_model.m_contour_points = m_object_model.m_prev_contour_points;
		return 2;
	}

	// Update the prev contour points
	m_object_model.m_prev_contour_points = m_object_model.m_contour_points;

	// Update the mask based on the current contour points. For fast level sets, the mask is automatically updated.
	if(m_contour_method != FAST_LEVEL_SETS) {
		Set(&m_object_mask, 0);
		FillPolygon(&m_object_mask, m_object_model.m_contour_points);
	}
	
#ifdef _OCCLUSSION_
	if(m_occlusion_handler.m_isObjectOccluded)
	{
		m_occlusion_handler.m_OccludedContour_points = m_object_model.m_contour_points;
		m_occlusion_handler.GetOccludedContourPoints(m_object_model, &m_object_model.m_contour_points, current_frame);
	}
#endif

	// UPDATE - if the object is not occluded
	if(!m_occlusion_handler.m_isObjectOccluded)
	{
		// Update the object attributes. If the object is occluded.. do not update anything.. return.
		if(!m_object_model.UpdateObjectAttributes(m_object_mask, dx, dy)) {
			//m_object_model.m_contour_points.clear();
			return 2;
		}

		
		// Update existing Fg and Bg spatial and appearance model
		m_object_model.UpdateFgSpatialModel(m_object_mask, dx, dy, m_use_region_consistency);
		m_object_model.UpdateFgAppearanceModel(current_frame);
		m_object_model.UpdateBgSpatialModel(m_object_mask);
		m_object_model.UpdateBgAppearanceModel(current_frame);
		

		// To update new modes in the appearance model of the fg
		//m_object_model.UpdateNewModes(current_frame, m_prev_frame, m_strength_image, m_object_mask);	
	}
	m_object_model.UpdateObjectHistory(frame_no, m_occlusion_handler.m_isObjectOccluded);

	// Handle occlusion
	bool flag = m_occlusion_handler.m_isObjectOccluded;
#ifdef _OCCLUSSION_
	if(m_frame_count > 5)
		m_occlusion_handler.HandleSingleObjectOcclussion(m_frame_count, m_object_model, current_frame);
	if(!flag && IsObjectOccluded)
		m_occlusion_handler.m_start_occ_frame_no = m_frame_no;
	if(flag && !IsObjectOccluded)
		m_occlusion_handler.m_end_occ_frame_no = m_frame_no;
#endif

	// Display the reconstructed scene.
	//_DisplayReconstructedScene();

	m_prev_frame = current_frame;
	return 1;
}

void ObjectTracker::_ComputeMotion(const ImgBgr& current_frame)
{
	int i;
	ImgBgr klt_display;
	ImgGray jklt_gray;
	Convert(current_frame, &jklt_gray);
	ImgBgr jklt_display;
	Convert(jklt_gray, &jklt_display);
	Convert(jklt_gray, &klt_display);
	if(m_use_motion) {
		ImgBinary region_mask;
		region_mask.Reset(m_width, m_height);
		float du = 0.0f, dv = 0.0f;
		if(m_frame_no >= 2) {
			for(i=0; i<m_object_model.m_fg_rgn_att.size(); i++) {
				if(m_object_model.m_fg_rgn_att[i].m_pixel_list.size() < 200)
					continue;
				Set(&region_mask, 0);
				for(list<Point>::iterator list_iter = m_object_model.m_fg_rgn_att[i].m_pixel_list.begin(); list_iter != m_object_model.m_fg_rgn_att[i].m_pixel_list.end(); list_iter++) {
					region_mask((*list_iter).x, (*list_iter).y) = true;
				}

				MotionDetector::ComputeJointKltMotion(m_prev_frame, current_frame, region_mask, m_object_model.m_fg_rgn_att[i].m_tracking_context, du, dv, jklt_display, m_nlevels, m_mindist, i);
		
				// update the spatial mean of each region based on the du and dv. Ideally the above procedure should be done for each region..
				m_object_model.m_fg_rgn_att[i].m_region_mean[0] += blepo_ex::Round(du);
				m_object_model.m_fg_rgn_att[i].m_region_mean[1] += blepo_ex::Round(dv);
			}
			m_object_model.m_object_att.m_region_mean[0] += blepo_ex::Round(du);
			m_object_model.m_object_att.m_region_mean[1] += blepo_ex::Round(dv);
		}

		static Figure motion_wnd("JKlt Motion");
		motion_wnd.Draw(jklt_display);

	}	// m_use_motion
}

void ObjectTracker::_DisplayReconstructedScene()
{
	int i;

	// Display the reconstructed object after update
	ImgGray region_test;
	region_test.Reset(m_width, m_height);
	Set(&region_test, 0);
	for(i=0; i<m_object_model.m_fg_rgn_att.size(); i++) {
		for(list<Point>::iterator list_iter = m_object_model.m_fg_rgn_att[i].m_pixel_list.begin(); list_iter != m_object_model.m_fg_rgn_att[i].m_pixel_list.end(); list_iter++) {
			Set(&region_test, m_object_model.m_fg_rgn_att[i].m_label, (*list_iter).x, (*list_iter).y, 1);
		}
	}

	ImgBgr region_disp;
	
	// Display the reconstructed background after update
	for(i=0; i<m_object_model.m_bkg_rgn_att.size(); i++) {
		if(m_object_model.m_bkg_rgn_att[i].m_is_occluded)
			continue;
		for(list<Point>::iterator list_iter = m_object_model.m_bkg_rgn_att[i].m_pixel_list.begin(); list_iter != m_object_model.m_bkg_rgn_att[i].m_pixel_list.end(); list_iter++) {
			region_test((*list_iter).x, (*list_iter).y) = m_object_model.m_bkg_rgn_att[i].m_label;
		}
	}
	Utils::RandomColor(region_test, &region_disp);
	static Figure bkg_wnd("Reconstructed scene");
	bkg_wnd.Draw(region_disp);

}

void ObjectTracker::DrawContour(ImgBgr &display_image)
{
	for(int i=0; i<m_object_model.m_contour_points.size(); i++) {
		DrawDot(m_object_model.m_contour_points[i], &display_image, m_contour_color, 2);
	}
}


void ObjectTracker::LoadStateFromFile(const ImgInt& current_model)
{
	int height = current_model.Height();
	int width = current_model.Width();
	int x, y, i;

	ImgBinary mask;
	mask.Reset(width, height);
	ImgBinary::Iterator p_mask = mask.Begin();
	ImgInt::ConstIterator p_model = current_model.Begin();

	int num_regions = m_object_model.m_fg_rgn_att.size();
	for(i=0; i<num_regions; i++) {
		m_object_model.m_fg_rgn_att[i].m_pixel_list.clear();
	}

	// Generate the mask
	for(y=0; y<height; y++) {
		for(x=0; x<width; x++, p_model++, p_mask++) {
			*p_mask = (*p_model > 0) ? true : false;
			for(i=0; i<num_regions; i++) {
				if(m_object_model.m_fg_rgn_att[i].m_label == *p_model) {
					m_object_model.m_fg_rgn_att[i].m_pixel_list.push_back(Point(x,y));
					break;
				}
			}
		}
	}

	Figure test_wnd("**8");
	test_wnd.Draw(mask);

	int dx, dy;
	m_object_model.UpdateObjectAttributes(mask, dx, dy);	
	m_object_model.UpdateFgSpatialModel(mask, dx, dy, false);
}

void ObjectTracker::_ComputeDecisionBoundary(const ImgBgr& current_frame)
{
	int x, y, i;

	double fg_strength, bg_strength, spatial_strength, max, bg_max;
	double total_fg_strength = 0.0f, total_bg_strength = 0.0f;
	int rgn_index, bg_rgn_index;
	double r_eps = 0.00000001f;  
	vector<int> fg_rgn_index;
	double feature_vec[3];
	double feature_strength;
	double t_spatial_strength;

	m_strength_image.Reset(256, 256);

	try {
	for(y=0; y<256; y++) {	
		for(x=0; x<256; x++) {
			fg_strength = 0.0f; bg_strength = 0.0f, spatial_strength = 0.0f; max = -1; bg_max = -1; rgn_index = -1; bg_rgn_index = -1; feature_strength = 0.0f; t_spatial_strength = 0.0f;
			fg_rgn_index.clear();
			for(i=0; i<m_object_model.m_fg_rgn_att.size(); i++) {
				feature_vec[0] = x; 	feature_vec[1] = y;  feature_vec[2] = 0;
				Utils::MahDistance3x3(feature_vec, m_object_model.m_fg_rgn_att[i].m_feature_mean, m_object_model.m_fg_rgn_att[i].m_inv_feature_cov, feature_strength);

				fg_strength += feature_strength; // * 10000.0f;
			}

			for(i=0; i<m_object_model.m_bkg_rgn_att.size(); i++) {					// for each bg region
				feature_vec[0] = x; feature_vec[1] = y; feature_vec[2] = 0;
				Utils::MahDistance3x3(feature_vec, m_object_model.m_bkg_rgn_att[i].m_feature_mean, m_object_model.m_bkg_rgn_att[i].m_inv_feature_cov, feature_strength);
				bg_strength += feature_strength;
			}
			
			// Calculate the log ratio
			m_strength_image(x,y) = log(blepo_ex::Max(fg_strength,r_eps )/blepo_ex::Max(bg_strength,r_eps));

			if(m_strength_image(x,y) <= 0)
				m_strength_image(x,y) = 0;
			else
				m_strength_image(x,y) = 255;
		}  // x
	}  // y

	 static Figure final_strength_wnd("Final Strength2", 0, m_height+20);
	 final_strength_wnd.Draw(m_strength_image);

	 ImgFloat grad_out;
	 ImgBinary canny_out;
	 ImgGray g_strength;

	 //GradMag(m_strength_image, 0.2f, &grad_out);
	 Convert(m_strength_image, &g_strength);
	 Canny(g_strength, &canny_out, 1.0f);
	 Figure grad_wnd("**9");
	 grad_wnd.Draw(canny_out);

	 FILE *fp = fopen("boundary.txt", "w");
	 for(y=0; y<256; y++) {
		 for(x=0; x<256; x++) {
			 if(canny_out(x,y))
				 fprintf(fp, "%d %d\n", x, y);
		 }
	 }
	 fclose(fp);

	 EventLoop();
	}
	catch(...) {
		cout << "Crashed in _ComputeStrengthMap" << endl;
	}

}

void ObjectTracker::_ComputeSpatialStrength()
{
	ImgFloat strength_map(m_width, m_height);
	int num_rgns = m_object_model.m_fg_rgn_att.size();
	double strength;
	int x, y, i;
	double spatial_vec[2];
	ImgBgr final_color_strength;
	final_color_strength.Reset(m_width, m_height);
	Set(&final_color_strength, Bgr(0,0,0));
	bool flag = false;

	// 3 - eye top, 2 - nose, 1 - body, 4 - eye small, 0 - mouth
	for(int k=0; k<num_rgns; k++) {

		if(k == 0) i = 1;
		else if(k == 1) i = 0;
		else if(k == 2) i = 2;
		else if(k == 3) i = 3;
		else if(k == 4) i = 4;
		//else if(k == 5) i = 4;

		Set(&strength_map, 0);

		for(y=0; y<m_height; y++) {
			for(x=0; x<m_width; x++) {
				spatial_vec[0] = x; spatial_vec[1] = y;
				Utils::MahDistance2x2(spatial_vec, m_object_model.m_fg_rgn_att[i].m_region_mean, m_object_model.m_fg_rgn_att[i].m_inv_region_cov, strength);

				strength_map(x,y) += strength;
			}
		}

		ImgFloat norm_strength;
		LinearlyScale(strength_map, 0, 1, &norm_strength);
		ImgBgr color_strength;
		int w = norm_strength.Width(); int h = norm_strength.Height();
		color_strength.Reset(w, h);
		Set(&color_strength, Bgr(255,255,255));
		for(y=0; y<h; y++) {
			for(x=0; x<w; x++) {
				float tmp = (norm_strength(x,y) < 0) ? 0 : norm_strength(x,y);
				if(tmp < 0.1f) continue;
				tmp += 0.2f;
				if(tmp > 1) tmp = 1;
				unsigned char val = tmp;
				if(i == 0)
					color_strength(x,y) = Bgr((int)0*tmp, (int)255*tmp, (int)0*tmp);
				else if(i == 1)
					color_strength(x,y) = Bgr((int)0*tmp, (int)0*tmp, (int)255*tmp);
				else if(i == 2)
					color_strength(x,y) = Bgr((int)0*tmp, (int)255*tmp, (int)255*tmp);
				else if(i == 3)
					color_strength(x,y) = Bgr((int)255*tmp, (int)0*tmp, (int)255*tmp);
				else if(i == 4)
					color_strength(x,y) = Bgr((int)255*tmp, (int)255*tmp, (int)0*tmp);
				else if(i == 5)
					color_strength(x,y) = Bgr((int)0*tmp, (int)255*tmp, (int)255*tmp);
			}
		}

		for(y=0; y<h; y++) {
			for(x=0; x<w; x++) {
				if(color_strength(x,y) != Bgr(255,255,255))
					final_color_strength(x,y) = color_strength(x,y);
			}
		}

		if(flag) break;
	}

	Figure region_strength_wnd("**11");
	region_strength_wnd.Draw(final_color_strength);
}
